-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local _G = getfenv(0)

local deformat = AceLibrary("Deformat-2.0")
local tablet = AceLibrary("Tablet-2.0")
local roster = AceLibrary("Roster-2.1")
local BC = AceLibrary("Babble-Class-2.2") 

local db = nil

local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = roster.roster[key] and roster.roster[key].class or "WTF"
		self[key] = "|cff" .. BC:GetHexColor(class) .. key .. "|r"
		return self[key]
	end
})

local dkpValues = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		self[key] = BananaDKP:GetDKP(key) or 0
		return self[key]
	end
})

local whisperHelp = nil
local whisperHandlers = nil
local commHandlers = {}
local commInitialized = nil

local recentWhispers = nil
local winner = nil

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

BananaDKP = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceConsole-2.0",
	"AceDB-2.0",
	"FuBarPlugin-2.0",
	"AceModuleCore-2.0",
	"AceHook-2.1"
)
local BananaDKP = BananaDKP

BananaDKP:SetModuleMixins("AceEvent-2.0")

local options = {
	type = "group",
	handler = BananaDKP,
	args = {
		dkpType = {
			type = "group",
			name = L["Type"],
			desc = L["Set what DKP system to use. Everyone must use the same system."],
			args = {},
			order = 101,
			disabled = function()
				return not BananaDKP:IsActive()
			end,
		},
		dkpTypeOptions = {
			type = "group",
			name = L["Type options"],
			desc = L["Options for the current DKP type."],
			args = {},
			order = 102,
			disabled = function()
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return not BananaDKP:IsActive()
			end,
		},
		role = {
			type = "group",
			name = L["Role"],
			desc = L["Sets your role, either client or server."],
			args = {
				client = {
					type = "toggle",
					name = L["Client"],
					desc = L["Sets this BananaDKP to act as a client."],
					get = function() return db.mode == 2 end,
					set = function(v)
						db.mode = v and 2 or 1
						BananaDKP:TriggerEvent("BananaDKP_ModeChanged", db.mode)
					end,
				},
				server = {
					type = "toggle",
					name = L["Server"],
					desc = L["Sets this BananaDKP to act as a server."],
					get = function() return db.mode == 1 end,
					set = function(v)
						db.mode = v and 1 or 2
						BananaDKP:TriggerEvent("BananaDKP_ModeChanged", db.mode)
					end,
				},
				host = {
					type = "text",
					name = L["Host"],
					desc = L["If acting as a client, set the host (player name) that is running the server here."],
					get = function() return db.host end,
					set = function(v)
						db.host = v
						BananaDKP:TriggerEvent("BananaDKP_HostChanged", db.host)
					end,
					usage = "<player>",
					disabled = function() return db.mode ~= 2 end,
				},
			},
			disabled = function()
				return not BananaDKP:IsActive()
			end,
		},
		sitouts = {
			type = "group",
			name = L["Sitouts"],
			desc = L["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."],
			disabled = function() return type(currentLeader) ~= "boolean" or not currentLeader end,
			args = {
				add = {
					type = "text",
					name = L["Add player"],
					desc = L["Adds a player to the sitout list."],
					get = false,
					usage = "<player name>",
					set = "AddSitoutPlayer",
					order = 1,
				},
				clear = {
					type = "execute",
					name = L["Clear list"],
					desc = L["Clear the sitout list."],
					func = "ClearSitoutList",
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 3,
				},
			},
			order = 200,
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return not BananaDKP:IsActive()
			end,
		},
	},
}

-----------------------------------------------------------------------
-- Temporary Debug Logging
-----------------------------------------------------------------------

local function Log(msg)
	if not BananaDKP.debugFrame then return end
	BananaDKP.debugFrame:AddMessage(msg)
end

-----------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------

-- Meh, can't set this in OnInitialize, since OnEmbedInitialize is called for
-- FBP before our OnInit, and FBP requires the property to be set then.
BananaDKP.hasIcon = "Interface\\Icons\\Ability_Hunter_AspectOfTheMonkey"

function BananaDKP:OnInitialize()
	local revision = tonumber((string.match("$Revision: 38922 $", "^.-(%d+).-$"))) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.BANANADKP_VERSION = self.version
	_G.BANANADKP_REVISION = self.revision

	self:RegisterDB("BananaDKPDB")
	self:RegisterDefaults("profile", {
		sitouts = {}, -- Remember this across sessions.
		dkpType = "Symphony", -- The DKP module to use.
		mode = 1, -- 1 = server, 2 = client
		host = nil, -- No host set by default.
		-- playerId = {}, -- Key: Name, Value: ID. Not using this, we're just
		-- using the player names for keys now - this should be fixed, or maybe
		-- an alias table could be added, for player alts that should have the
		-- same DKP.
	})

	self.OnMenuRequest = options

	self:RegisterChatCommand("/bdkp", options, "BANANADKP")

	self.debugFrame = ChatFrame1

	db = self.db.profile
end

function BananaDKP:OnEnable(first)
	local currentModule = db.dkpType
	for name, module in self:IterateModules() do
		if name == currentModule then
			self:SetActiveDKPModule(name)
			break
		end
	end

	-- For hiding DKP whispers.
	self:Hook("ChatFrame_MessageEventHandler", "WhisperHandler", true)

	self:RegisterEvent("BananaDKP_DKPUpdated")
	self:RegisterEvent("CHAT_MSG_LOOT")

	self:RebuildSitoutList()
end

function BananaDKP:OnDisable()
	for name, module in self:IterateModules() do
		self:ToggleModuleActive(module, false)
	end
end

-----------------------------------------------------------------------
-- Events
-----------------------------------------------------------------------

-- Someone thinks we should be synchronizing.
function BananaDKP:BananaDKP_DKPUpdated()
	-- We get this event every time the DKP table changes, so clear our cache.
	for k, v in pairs(dkpValues) do
		dkpValues[k] = nil
	end
end

function BananaDKP:CHAT_MSG_LOOT(msg)
	local player, item = deformat(msg, LOOT_ITEM)
	if not player then
		item = deformat(msg, LOOT_ITEM_SELF)
		if item then
			player = UnitName("player")
		end
	end

	if type(item) == "string" and type(player) == "string" then
		local itemName, itemLink, itemRarity = GetItemInfo(item)
		if itemName == "Void Crystal" or itemName == "Badge of Justice" then return end
		-- Unconditionally notify modules regardless of itemRarity. Of course,
		-- most times modules will only care if itemRarity is higher than 3
		-- (epic or higher), but you never know.
		if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
			for name, module in self:IterateModules() do
				if self:IsModuleActive(name) and type(module.ItemLooted) == "function" then
					module:ItemLooted(player, itemLink, itemRarity)
					break
				end
			end
		end
	end
end

-----------------------------------------------------------------------
-- Hooks
-----------------------------------------------------------------------

function BananaDKP:RegisterWhisperHandler(prefix, handlerFunc, prefixHelp)
	assert(type(prefix) == "string", "Invalid prefix.")
	prefix = prefix:lower()
	assert(not whisperHandlers[prefix], "Prefix already registered.")

	whisperHelp[prefix] = prefixHelp
	whisperHandlers[prefix] = handlerFunc
end

local function tablePrinter(t, recipient)
	for i, v in ipairs(t) do
		if type(v) == "string" then
			BananaDKP:Chat("BDKP: " .. v, "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

function BananaDKP:WhisperHandler(event, ...)
	if db.mode == 2 and roster.roster[db.host] then return self.hooks["ChatFrame_MessageEventHandler"](event, ...) end
	if (event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM") and type(arg1) == "string" and type(arg2) == "string" then
		local text = arg1:lower()
		if text:find("^b.-dkp") then
			if event == "CHAT_MSG_WHISPER_INFORM" then
				return
			end
			local player = arg2:lower()
			if type(recentWhispers) ~= "table" then
				recentWhispers = {}
			end
			local t = recentWhispers[player]
			if not t or (GetTime() - t > 1) then
				recentWhispers[player] = GetTime()
				local command, rest = select(3, text:find("^b.-dkp (%w+)(.*)"))
				if command then
					command = command:gsub("%A", ""):lower():trim()  -- So we don't disconnect if someone tries to send us item links.
				end
				if rest then
					rest = rest:gsub("%A", ""):lower():trim()
					if rest:len() < 3 then rest = nil end
				end
				if command and type(whisperHandlers[command]) == "function" then
					local ret = whisperHandlers[command](event, rest or arg2:lower())
					if type(ret) == "string" and ret:trim():len() > 3 then
						self:Chat(string.format(L["BananaDKP: %s."], ret), "WHISPER", nil, arg2)
					elseif type(ret) == "table" then
						tablePrinter(ret, arg2)
					end
				elseif command and command == "help" then
					local ret = self:GetHelpText(rest)
					if type(ret) == "string" then
						self:Chat(ret, "WHISPER", nil, arg2)
					elseif type(ret) == "table" then
						for i, v in ipairs(ret) do
							self:Chat(v, "WHISPER", nil, arg2)
						end
					end
				else
					player = command or arg2
					local dkp = dkpValues[player]
					if type(dkp) == "number" then
						self:Chat(string.format(L["BananaDKP: DKP for %s: %s."], player, tostring(dkp)), "WHISPER", nil, arg2)
					else
						self:Chat(L["BananaDKP: Unknown command or player not found."], "WHISPER", nil, arg2)
					end
				end
				return
			end
		elseif currentStatus == STATUS_TAKING_BIDS and text:find("^bid") then
			self:NewBidder(arg2)
		end
	elseif event == "CHAT_MSG_RAID" or event == "CHAT_MSG_RAID_LEADER" and type(arg2) == "string" and type(currentLeader) == "boolean" and currentLeader == true then
		if arg2 == UnitName("player") and arg1:lower():find(L["^bid:"]) then
			if currentStatus ~= STATUS_TAKING_BIDS then
				currentStatus = STATUS_TAKING_BIDS
				self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
			else
				-- We're already taking bids, but someone won it and we're going
				-- again with a new item.
				winner = nil
				self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
			end
		elseif currentStatus == STATUS_TAKING_BIDS and arg1:lower():find(L["^bid"]) then
			self:NewBidder(arg2)
		end
	end
	return self.hooks["ChatFrame_MessageEventHandler"](event, ...)
end

function BananaDKP:NewBidder(bidder)
	local dkp = dkpValues[arg2]
	if not winner or winner == bidder then
		winner = bidder
	else
		local winnerDkp
		if type(winner) == "string" then
			winnerDkp = dkpValues[winner]
		elseif type(winner) == "table" then
			winnerDkp = dkpValues[winner[1]]
		end
		if not winnerDkp then winnerDkp = -100 end
		if dkp > winnerDkp then
			winner = bidder
		elseif dkp == winnerDkp then
			if type(winner) ~= "table" then
				winner = {winner}
			end
			local skip = nil
			for i, v in ipairs(winner) do
				if v == bidder then
					skip = true
				end
			end
			if not skip then
				table.insert(winner, bidder)
			end
		end
	end
	self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
end

function BananaDKP:GetHelpText(topic)
	if topic then
		if whisperHelp[topic] then
			return whisperHelp[topic]
		elseif topic == "bid" then
			return L["Chat-Help-Bid"]
		elseif topic == "dkp" then
			return L["Chat-Help-Dkp"]
		end
	else
		local help = {}
		table.insert(help, L["Chat-Help-Dkp"])
		table.insert(help, L["Chat-Help-Bid"])
		for k, v in pairs(whisperHelp) do
			table.insert(help, v)
		end
		return help
	end
end

-----------------------------------------------------------------------
-- Communication Handling
-----------------------------------------------------------------------

function BananaDKP:Chat(text, chatType, language, destination)
	if ChatThrottleLib then
		ChatThrottleLib:SendChatMessage("BULK", "BananaDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

-----------------------------------------------------------------------
-- Communication Timeout Handling
-----------------------------------------------------------------------

function BananaDKP:BidTimeout()
	if type(winner) == "table" then
		self:Chat(string.format(L["Tie: %s."], table.concat(winner, ", ")), "RAID")
	elseif type(winner) == "string" then
		self:Chat(string.format(L["%s wins!"], tostring(winner)), "RAID")
	end
	winner = nil
end

-----------------------------------------------------------------------
-- FuBar Handling
-----------------------------------------------------------------------

function BananaDKP:OnTextUpdate()
	self:SetText(db.mode == 1 and L["Server"] or L["Client"])
end

function BananaDKP:OnTooltipUpdate()
	local cat = tablet:AddCategory(
		"columns", 2,
		"child_textR", 1,
		"child_textG", 1,
		"child_textB", 0,
		"child_text2R", 1,
		"child_text2G", 1,
		"child_text2B", 1
	)
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			cat:AddLine("text", L["Type"], "text2", name)
			break
		end
	end
	cat = tablet:AddCategory(
		"columns", 2,
		"text", L["Player"],
		"text2", L["DKP"],
		"child_textR", 1,
		"child_textG", 1,
		"child_textB", 0,
		"child_text2R", 1,
		"child_text2G", 1,
		"child_text2B", 1
	)
	for n, u in pairs(roster.roster) do
		if u.class ~= "PET" then
			cat:AddLine("text", coloredName[n], "text2", dkpValues[n] or 0)
		end
	end
end

-----------------------------------------------------------------------
-- DKP Module Handling
-----------------------------------------------------------------------

function BananaDKP:GetMode()
	return db.mode
end

function BananaDKP:GetHost()
	return db.host
end

function BananaDKP:GetDKP(player)
	local dkp = nil
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			dkp = module:GetDKP(player)
			break
		end
	end
	return dkp
end

function BananaDKP:SetActiveDKPModule(moduleName)
	assert(self:HasModule(moduleName), moduleName .. " doesn't seem to exist.")
	assert(not running, "Can't change the DKP type while active.")

	db.dkpType = moduleName

	-- Clear the registered whisper handlers.
	whisperHandlers = {}
	whisperHelp = {}

	for name, module in self:IterateModules() do
		if name == moduleName then
			self:ToggleModuleActive(module, true)
			options.args.dkpTypeOptions.args = module.options
		else
			self:ToggleModuleActive(module, false)
		end
	end

	self:UpdateDisplay()
end

function BananaDKP:RegisterDKPType(name, module)
	if module:IsRegistered() then
		error(name .. " has already been registered.")
	end
	if type(module.defaultDatabase) ~= "table" then
		error(name .. " needs a default database.")
	end
	if type(module.defaultDatabase.databaseVersion) ~= "number" then
		error(name .. "'s database needs a databaseVersion number field.")
	end
	if type(module.GetDKP) ~= "function" then
		error(name .. " needs a GetDKP method.")
	end
	if type(module.GetLootedItems) ~= "function" then
		error(name .. " needs a GetLootedItems method.")
	end

	module.db = self:AcquireDBNamespace(name)
	self:RegisterDefaults(name, "profile", module.defaultDatabase)

	self:ToggleModuleActive(module, false)

	local opt = options.args.dkpType.args

	opt[name] = {
		type = "toggle",
		name = name,
		desc = string.format(L["Activate %s."], name),
		get = function() return BananaDKP.db.profile.dkpType == name end,
		set = function()
			BananaDKP:SetActiveDKPModule(name)
		end,
		disabled = function() return running end,
	}

	if type(module.OnRegister) == "function" then
		module:OnRegister()
	end
end

-----------------------------------------------------------------------
-- DKP Type prototype
-----------------------------------------------------------------------

BananaDKP.modulePrototype.core = BananaDKP

function BananaDKP.modulePrototype:OnInitialize()
	assert(not self.registered, self.name .. " has already been registered.")
	self.core:RegisterDKPType(self.name, self)
end

function BananaDKP.modulePrototype:IsRegistered()
	return self.registered
end

-----------------------------------------------------------------------
-- Sitout list handling
-----------------------------------------------------------------------

function BananaDKP:ClearSitoutList()
	db.sitouts = {}
	self:RebuildSitoutList()
end

function BananaDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	self:RebuildSitoutList()
end

function BananaDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	self:RebuildSitoutList()
end

function BananaDKP:RebuildSitoutList()
	local opts = options.args.sitouts.args

	for k, v in pairs(opts) do
		if k ~= "add" and k ~= "clear" and k ~= "spacer" then
			opts[k] = nil
		end
	end

	for k, v in pairs(db.sitouts) do
		opts[k] = {
			type = "execute",
			name = k,
			desc = string.format(L["Remove %s."], k),
			func = function() BananaDKP:RemoveSitoutPlayer(k) end,
			order = 100,
		}
	end
end

